#include <iostream.h>

#include <qapp.h>
#include <qclipbrd.h>
#include <qdragobject.h> 
#include <qkeycode.h>
#include <qmessagebox.h> 
#include <kapp.h>
#include "hexviewwidget.h"




//
// This widget will use the entire space of the parent widget
//
CHexViewWidget::CHexViewWidget( QWidget *parent, CHexBuffer *hexBuffer )
  : QWidget( parent )
{
  if( parent == 0 || hexBuffer == 0 ) { return; }

  mVertScroll = new QScrollBar( QScrollBar::Vertical, parent );
  if( mVertScroll == 0 ) { return; }
  mHorzScroll = new QScrollBar( QScrollBar::Horizontal, parent );
  if( mHorzScroll == 0 ) { return; }
  mSquare = new QWidget( parent );
  if( mSquare == 0 ) { return; }



  connect( mHorzScroll, SIGNAL(valueChanged(int)), SLOT(changeXPos(int)) );
  connect( mVertScroll, SIGNAL(valueChanged(int)), SLOT(changeYPos(int)) );
  connect( kapp, SIGNAL( kdisplayPaletteChanged() ),SLOT( paletteChanged() ) );
  connect( kapp, SIGNAL( kdisplayFontChanged() ), SLOT( fontChanged() ) );

  mHexBuffer = hexBuffer;
  mHexBuffer->cursorReset();
  mHexBuffer->setSelectionColor( kapp->selectTextColor, kapp->selectColor );
  mHexBuffer->setBackgroundColor( kapp->windowColor );
  mHexBuffer->setFont( kapp->fixedFont );

  mScrollBarSize = 16;
  mEditMode      = CHexBuffer::edit_overwrite;
  mFixedSize     = false;
  mReplacePrompt = 0;
  mCursorTimerId = 0;
  mShowCursor    = false;
  mDisableCursor = false;

  setTextBufferSize(); // Make sure there is a pixmap buffer
  setStartX(0);
  setStartY(0);
  setCursorBlinkInterval( 500 );

  setBackgroundColor( hexBuffer->textBackgroundColor() );
  setFocusPolicy( StrongFocus );
}

CHexViewWidget::~CHexViewWidget( void )
{
  delete mVertScroll;
  delete mHorzScroll;
  delete mSquare;
  delete mReplacePrompt;
}


void CHexViewWidget::loadFile( QFile *file, const QString &url, bool insert ) 
{
  if( insert == true && mHexBuffer->isValid() == true )
  {
    uint cursorLine = mHexBuffer->cursorLine();
    bool success = mHexBuffer->insertFile( file, url );
    if( success == true )
    {
      updateWindow( cursorLine );
      emit cursorChanged( mHexBuffer->cursorState() );
      emit fileModified( mHexBuffer->fileState() );
      emit displayChanged( mHexBuffer->displayState() );
    }
  }
  else
  {
    bool success = mHexBuffer->loadFile( file, url );
    if( success == true )
    {
      initFile();
    }
  }
}


void CHexViewWidget::closeFile( void ) 
{
  emit fileClosed( mHexBuffer->url() );
  mHexBuffer->closeFile();
  initFile();
}



void CHexViewWidget::initFile( void )
{
  setStartX(0);
  setStartY(0);

  setEditMode( mEditMode );
  setBackgroundMode( NoBackground ); // Prevents flicker

  mHexBuffer->setDisableCursor( mDisableCursor );
  mHexBuffer->cursorReset();

  updateView( true, false );
  resizeEvent( 0 );

  emit cursorChanged( mHexBuffer->cursorState() );
  emit fileModified( mHexBuffer->fileState() );
  emit fileName( mHexBuffer->url() );
}


void CHexViewWidget::setBuffer( CHexBuffer *hexBuffer )
{
  if( hexBuffer == 0 )
  {
    return;
  }

  mHexBuffer = hexBuffer;

  mHexBuffer->setDisplayState( mDpyState );
  
  if( mFixedSize == false )
  {
    mHexBuffer->matchWidth( width() );
  }

  mHexBuffer->setSelectionColor( kapp->selectTextColor, kapp->selectColor );
  mHexBuffer->setBackgroundColor( kapp->windowColor );
  mHexBuffer->setFont( kapp->fixedFont );
  mHexBuffer->setDisableCursor( mDisableCursor );
  setBackgroundColor( hexBuffer->textBackgroundColor() );
  setBackgroundMode( NoBackground ); // Prevents flicker


  setEditMode( mEditMode );
  updateWindow();

  emit cursorChanged( mHexBuffer->cursorState() );
  emit fileModified( mHexBuffer->fileState() );
  emit displayChanged( mHexBuffer->displayState() );
  emit fileName( mHexBuffer->url() );
}




void CHexViewWidget::changeXPos( int p ) 
{ 
  int dx = startX() - p;
  setStartX(p);


  if( QABS(dx) < width() )
  {
    scroll( dx, 0 );
  }
  else
  {
    QWidget::update();
  }

  //
  // If the start position has become 0, then update the view. This
  // will remove the scrollbar (if it is visible) if the textarea width
  // is wider than the text. The scrollbar will then disappear under the 
  // mouse pointer.
  //
  if( startX() == 0 ) 
  { 
    updateView( false, false ); 
  }

}


void CHexViewWidget::changeYPos( int p ) 
{
  int dy  = startY() - p;
  setStartY(p);

  if( QABS( dy ) < height() )
  {
    scroll( 0, dy );
  }
  else
  {
    QWidget::update();
  }

  //
  // If the start position has become 0, then update the view. This
  // will remove the scrollbar (if it is visible) if the textarea height
  // is taller than the text. The scrollbar will then disappear under the 
  // mouse pointer.
  //
  if( startY() == 0 )
  {
    updateView( false, false );
  }
}


void CHexViewWidget::clipboardChanged( void )
{
  disconnect(QApplication::clipboard(),SIGNAL(dataChanged()),
	     this,SLOT(clipboardChanged()));
  unselect( 0 );
}

void CHexViewWidget::paletteChanged( void )
{
  mHexBuffer->setSelectionColor( kapp->selectTextColor, kapp->selectColor );
  mHexBuffer->setBackgroundColor( kapp->windowColor );
  repaint();
}

void CHexViewWidget::fontChanged( void )
{
  mHexBuffer->setFont( kapp->fixedFont );
  emit textWidth( mHexBuffer->lineWidth() + scrollBarWidth() );
  updateWindow();
}


void CHexViewWidget::insert( SInsertData &id )
{
  if( id.onCursor == false )
  {
    mHexBuffer->cursorGoto( id.offset, 0 );
  }
  SCursorConfig cc;
  updateCursor( cc, true ); 

  if( id.size == 0 )
  {
    return;
  }

  char *buf = new char [ id.size ];
  if( buf == 0 )
  {
    return;
  }
  memset( buf, 0, id.size );
  if( id.patternSize > 0 )
  {
    uint size = id.patternSize > id.size ? id.size : id.patternSize;
    if( id.repeatPattern == false )
    {
      memcpy( buf, id.pattern, size );
      if( size < id.size )
      {
	memset( &buf[size], id.pattern[id.patternSize-1], id.size - size );
      }
    }
    else
    {
      for( uint i=0; i < id.size; i+= size )
      {
	uint s = i+size > id.size ? id.size-i : size;
	memcpy( &buf[i], id.pattern, s );
      }
    }
  }

  uint cursorLine = mHexBuffer->cursorLine();

  bool success = mHexBuffer->inputAtCursor( buf, id.size, 0 );
  if( success == true )
  {
    updateWindow( cursorLine );
  }

  delete [] buf;
}




void CHexViewWidget::updateView( bool redraw, bool fixCursor )
{
  for( uint i=0; i < 2; i++ )
  {
    int  editWidth    = width();
    int  editHeight   = height();
    int  textWidth    = mHexBuffer->lineWidth();
    int  textHeight   = mHexBuffer->totalHeight();

    //
    // This will move the start position (horizontal) to the left
    // (if possible) if the text width is smaller than the edit width.
    //
    int size = mVertScroll->isVisible() == true ? mScrollBarSize : 0;
    if( startX() + editWidth - size > textWidth )
    {
      int position = textWidth - editWidth + size;
      setStartX( position > 0 ? position : 0 );
    }

    int xMax = textWidth - editWidth;
    bool horzScrollbarVisible = startX() > 0 || xMax > 0 ? true : false;
    if( horzScrollbarVisible == true ) 
    { 
      editHeight -= mScrollBarSize; 
    }

    //
    // This will move the start position (vertical) to the top
    // (if possible) if the text height is smaller than the edit height.
    //
    if( startY() + editHeight > textHeight )
    {
      int position = textHeight - editHeight;
      if( position >= 0 )
      {
	setStartY( position );
      }
    }


    int yMax      = textHeight - editHeight;
    int startLine = startY() / textHeight;

    if( startLine > 0 || yMax > 0 ) 
    {
      editWidth -= mScrollBarSize;
      xMax += mScrollBarSize;
      if( horzScrollbarVisible == false && xMax > 0 ) 
      {
	// Horizontal scrollbar will be visible after all.
	editHeight -= mScrollBarSize;
	yMax += mScrollBarSize;
      }
    }
    
    if( xMax < startX() ) { xMax = startX(); }
    if( yMax < startY() ) { yMax = startY(); }

    if( xMax > 0 ) 
    {
      mHorzScroll->blockSignals( true );
      mHorzScroll->setGeometry( 0, editHeight, editWidth, mScrollBarSize );
      mHorzScroll->setRange( 0, xMax );
      mHorzScroll->setValue( startX() );
      mHorzScroll->setSteps( mHexBuffer->lineHeight(),
			     editWidth-mScrollBarSize );
      mHorzScroll->blockSignals( false );
      mHorzScroll->show();
    } 
    else 
    {
      mHorzScroll->hide();
    }

    if( yMax > 0 ) 
    {
      mVertScroll->blockSignals( true );
      mVertScroll->setGeometry( editWidth, 0, mScrollBarSize, editHeight );
      mVertScroll->setRange( 0, yMax );
      mVertScroll->setValue( startY() );
      mVertScroll->setSteps( mHexBuffer->lineHeight(),
			     editHeight-mScrollBarSize );
      mVertScroll->blockSignals( false );
      mVertScroll->show();
    } 
    else 
    {
      mVertScroll->hide();
    }

    if( mHorzScroll->isVisible() == true && mVertScroll->isVisible() == true )
    {
      mSquare->setGeometry( editWidth, editHeight, mScrollBarSize, 
			    mScrollBarSize );
      mSquare->show();
    }
    else
    {
      mSquare->hide();
    }
    
    if( fixCursor == true )
    {
      int position = mHexBuffer->cursorFixedPosition( startY(), height() );
      if( position != startY() )
      {
	setStartY( position );
	fixCursor = false;
	continue;
      }
    }

    break;
  }

  if( redraw == true )
  {
    QWidget::update();
  }

}



void CHexViewWidget::setDisplayState( SDisplayState &state )
{
  mDpyState = state;
  mHexBuffer->setDisplayState( mDpyState );
  updateView( true/*false*/, true );
  emit displayChanged( mHexBuffer->displayState() );
  emit cursorChanged( mHexBuffer->cursorState() );
  emit textWidth( mHexBuffer->lineWidth() + scrollBarWidth() );
};


void CHexViewWidget::setCursorBlinkInterval( uint interval )
{
  mCursorBlinkInterval = interval;
  if( mDisableCursor == false )
  {
    enableCursorTimer( true );
    paintCursor( CHexBuffer::cursor_curr );
  }
}



void CHexViewWidget::selectAll( void )
{
  setSelection( 0, true );
  setSelection( mHexBuffer->textSize(), false );
  copy();
  emit cursorChanged( mHexBuffer->cursorState() );
}

void CHexViewWidget::unselect( uint offset )
{
  setSelection( offset, true );
  emit cursorChanged( mHexBuffer->cursorState() );
}


void CHexViewWidget::search( USearchData &sd )
{
  bool success = mHexBuffer->search( sd, false );
  if( success == false )
  {
    return;
  }

  updateWindow( true, false );

  if( sd.mAny.operation == USearchData::opReplace )
  {
    mReplaceData = sd;
    mReplaceData.mReplace.fromCursor = true;
    mNumReplaced = 0;

    if( sd.mReplace.prompt == true )
    {
      createReplacePrompt( sd );
    }
    else
    {
      replacePromptAction( repAll );
    }
  }
}



void CHexViewWidget::createReplacePrompt( USearchData &sd )
{
  if( mReplacePrompt == 0 )
  {
    mReplacePrompt = new CReplacePromptDialog( 0, "replace" );
    if( mReplacePrompt == 0 ) { return; }
    mReplacePrompt->setCaption( i18n("Find and Replace") );

    connect( mReplacePrompt, SIGNAL(execute( int )), 
	     this, SLOT(replacePromptAction( int )) );

  }
  if( mReplacePrompt->isVisible() == false )
  {
    centerDialog( mReplacePrompt, topLevelWidget() );
    mReplacePrompt->show();
  }
  else
  {
    mReplacePrompt->raise();
  }
}


void CHexViewWidget::finishReplacePrompt( bool showStatistics )
{
  emit fileModified( mHexBuffer->fileState() );

  if( mReplacePrompt != 0 )
  {
    mReplacePrompt->hide();
  }

  if( showStatistics == true )
  {
    QString msg;
    msg += i18n( "Operation complete" );
    msg += "\n\n";

    if( mNumReplaced == 0 )
    {
      QString msg2;
      msg2.sprintf( "%s.", i18n( "No replacements were made" ) );
      msg += msg2;
    }
    else if( mNumReplaced == 1 )
    {
      QString msg2;
      msg2.sprintf( "%s.", i18n( "One replacement was made" ) );
      msg += msg2;
    }
    else
    {
      QString msg2;
      msg2.sprintf( "%u %s.", mNumReplaced, i18n( "replacements were made" ) );
      msg += msg2;
    }

    QMessageBox::information(
      this ,
      i18n("Find and Replace"),
      msg,
      i18n( "OK" ),
      "",
      "",
      0, 0 );
  }

}


void CHexViewWidget::replacePromptAction( int retCode )
{
  switch( retCode )
  {
    case repYes:
    { 
      mReplaceData.mReplace.step = false;
      bool success = mHexBuffer->search( mReplaceData, true );
      if( success == false )
      {
	finishReplacePrompt( false );
	return;
      }

      mNumReplaced += 1;
      updateWindow( true, false );

      success = mHexBuffer->search( mReplaceData, false );
      if( success == false )
      {
	finishReplacePrompt( true );
	return;
      }

      updateWindow( true, false );

    }
    break;

    case repNo:
    {
      mReplaceData.mReplace.step = true;
      bool success = mHexBuffer->search( mReplaceData, false );
      if( success == false )
      {
	finishReplacePrompt( true );
	return;
      }

      updateWindow( true, false );
    }
    break;

    case repAll:
    {
      mNumReplaced += mHexBuffer->replaceAll( mReplaceData );
      updateWindow( true, false );
      finishReplacePrompt( true );
      return;
    }
    break;

    case repClose:
    break;
  }
}


bool CHexViewWidget::export( QByteArray &array, SExportData &ed )
{
  return( mHexBuffer->export( array, ed ) );
}


void CHexViewWidget::copy( void )
{
  QByteArray raw;
  if( mHexBuffer->selectedData( raw, false ) == false )
  {
    return;
  }

  QByteArray encoded;
  bool success = mClipConvert.encode( encoded, raw );
  if( success == false )
  {
    return;
  }

  disconnect(QApplication::clipboard(),SIGNAL(dataChanged()),
	     this,SLOT(clipboardChanged()));
  QApplication::clipboard()->setText( encoded.data() );
  connect(QApplication::clipboard(),SIGNAL(dataChanged()),
	  this,SLOT(clipboardChanged()));
}


void CHexViewWidget::copyFormatted( bool allField )
{
  SExportData ed;
  ed.mode = SExportData::SelectedArea;

  QByteArray array;
  if( mHexBuffer->export( array, ed ) == false )
  {
    return;
  }

  disconnect(QApplication::clipboard(),SIGNAL(dataChanged()),
	     this,SLOT(clipboardChanged()));
  QApplication::clipboard()->setText( array.data() );
  connect(QApplication::clipboard(),SIGNAL(dataChanged()),
	  this,SLOT(clipboardChanged()));
}


void CHexViewWidget::paste( void )
{
  QString buf = QApplication::clipboard()->text();
  if( buf.isEmpty() == true )
  {
    return;
  }

  QByteArray decoded;
  bool success = mClipConvert.decode( decoded, buf );
  if( success == false )
  {
    return;
  }

  success = mHexBuffer->inputAtCursor( decoded.data(), decoded.size(), 0 );
  if( success == false )
  {
    return;
  }

  updateWindow( true, true );
}


void CHexViewWidget::cut( void )
{
  copy();

  bool success = mHexBuffer->cutSelection();
  if( success == false )
  {
    return;
  }

  updateWindow( false, true );
}


void CHexViewWidget::undo( void )
{
  bool success = mHexBuffer->undo();
  if( success == false )
  {
    return;
  }
 
  updateWindow( true, true );
}


void CHexViewWidget::redo( void )
{
  bool success = mHexBuffer->redo();
  if( success == false )
  {
    return;
  }

  updateWindow( true, true );
}



#include <sys/time.h>
#include <unistd.h>

void CHexViewWidget::benchmark( void )
{
  struct timeval t1, t2;
  QRect r( 0, 0, width(), height() );

  uint loop = 10;

  gettimeofday( &t1, 0 );
  for( uint i=0; i< loop; i++ )
  {
    paintText( r );
  }
  gettimeofday( &t2, 0 );


  uint area  = width() * height();
  uint last = (t2.tv_sec-t1.tv_sec) * 1000000 + (t2.tv_usec - t1.tv_usec);
  
  debug( "Duration:      %f", (float)last/ 1000000.0 );
  debug( "Duration/loop: %f", (float)last/ (1000000.0*(float)loop) );
  debug( "Area:          %d", area );
  debug( "Loop:          %d", loop );
}





void CHexViewWidget::resizeEvent( QResizeEvent *e )
{
  setTextBufferSize();

  if( mFixedSize == true )
  {
    updateView( false, false );
  }
  else
  {
    bool state = mVertScroll->isVisible();
    int size = state == true ? mScrollBarSize : 0;

    bool bufferChanged = mHexBuffer->matchWidth( width() - size );
    updateView( false, bufferChanged );
  
    if( mVertScroll->isVisible() != state )
    {
      size = mVertScroll->isVisible() ? mScrollBarSize : 0;
      bufferChanged = mHexBuffer->matchWidth( width() - size );
      updateView( false, bufferChanged );
    }
  }

}


void CHexViewWidget::paintEvent( QPaintEvent *e )
{
  paintText( e->rect() );
}



void CHexViewWidget::keyPressEvent( QKeyEvent *e )  
{
  SCursorConfig cc;
  cc.state = e->state();

  //
  // Some special actions that we have to trap here
  //
  if( e->state() & ControlButton )
  {
    switch( e->key() )
    {
      case Key_Space:
	e->accept();
	stepEditArea();
	return;
      break;

      case Key_1:
	e->accept();
	cursorStep( cc, 1 );
	return;
      break;

      case Key_2:
	e->accept();
	cursorStep( cc, 2 );
	return;
      break;

      case Key_4:
	e->accept();
	cursorStep( cc, 4 );
	return;
      break;

      case Key_8:
	e->accept();
	cursorStep( cc, 8 );
	return;
      break;
    }
  }

  if( e->state() & AltButton ) 
  {
    e->ignore();
    return;
  }

  switch ( e->key() ) 
  {
    case Key_Left:
      cursorLeft( cc );
    break;
    
    case Key_Right:
      cursorRight( cc );
    break;

    case Key_Up:
      cursorUp( cc );
    break;

    case Key_Down:
      cursorDown( cc );
    break;

    case Key_Home:
      cursorHome( cc );
    break;
    
    case Key_End:
      cursorEnd( cc );
    break;

    case Key_Next:
      cursorPageDown( cc );
    break;

    case Key_Prior:
      cursorPageUp( cc );
    break;

    case Key_Insert:
      cursorInsert( cc );
    break;

    case Key_Delete:
      cursorDelete( cc );
    break;

    case Key_Backspace:
      cursorBackspace( cc );
    break;

    default:
      if( e->ascii() > 0 )  // Is 0 when modifier only is pressed
      {
	cursorInput( e->ascii() );
      }
    break;
  }

  e->accept();
}

void CHexViewWidget::keyReleaseEvent( QKeyEvent *e )
{
  if( e->state() & ShiftButton && shiftButtonState() == false )
  {
    //
    // The shift button was pressed when event was triggered, but is 
    // now released. I use this as a sign to copy selected data to the 
    // clipboard. The shiftButtonState() uses X-lib functions so have a 
    // look in the headerfile (inlined method) if you want to remove
    // this dependency.
    //
    copy();
  }
}





void CHexViewWidget::mousePressEvent( QMouseEvent *e ) 
{
  if( e->button() == LeftButton )
  {
    setCursor( e->x(), e->y(), true );
  }
  else if( e->button() == RightButton )
  {
  }
  else if( e->button() == MidButton )
  {
    if( mHexBuffer->isValid() == false )
    {
      emit openNewFile();
    }
    paste();
  }

}

void CHexViewWidget::mouseMoveEvent( QMouseEvent *e ) 
{
  if( e->state() & LeftButton ) 
  {
    setCursor( e->x(), e->y(), false );
  }
}

void CHexViewWidget::mouseReleaseEvent( QMouseEvent *e ) 
{
  mHexBuffer->cursorResetEditArea();
  copy();
}


void CHexViewWidget::timerEvent( QTimerEvent *e )
{
  if( mCursorBlinkInterval == 0 )
  {
    mShowCursor = true;
  }
  else
  {
    mShowCursor = mShowCursor == true ? false : true;
  }
  mHexBuffer->setShowCursor( mShowCursor );
  paintCursor( CHexBuffer::cursor_curr );
}

void CHexViewWidget::focusInEvent( QFocusEvent *e )
{
  mDisableCursor = false;
  mHexBuffer->setDisableCursor( mDisableCursor );
  enableCursorTimer( true );
  paintCursor( CHexBuffer::cursor_curr );
}

void CHexViewWidget::focusOutEvent( QFocusEvent *e )
{
  mDisableCursor = true;
  mHexBuffer->setDisableCursor( mDisableCursor );
  paintCursor( CHexBuffer::cursor_curr );
}


void CHexViewWidget::setSelection( uint offset, bool init )
{
  bool selectionChanged = mHexBuffer->selectionSet( offset, init );
  if( selectionChanged == true )
  {
    uint off1, off2;
    mHexBuffer->selectionStartChange( off1, off2 );
    if( off1 != off2 )
    {
      paintText( off1, off2 );
    }
    mHexBuffer->selectionStopChange( off1, off2 );
    if( off1 != off2 )
    {
      paintText( off1, off2 );
    }
  }
  mHexBuffer->selectionSyncronize();
}



void CHexViewWidget::setCursor( int x, int y, bool init )
{
  int editArea = mHexBuffer->setCursor( x + startX(), y + startY(), init );
  if( editArea == CHexBuffer::edit_none )
  {
    if( init == true ) { setSelection( 0, true ); }
    return;
  }

  SCursorConfig cc;
  cc.setKeepSelection( init == true ? false : true );
  updateCursor( cc, init );
}



void CHexViewWidget::paintText( uint start, uint stop )
{
  QPainter paint;
  paint.begin( &mTextBuffer );
  paint.setFont( mHexBuffer->font() );

  uint lineHeight = mHexBuffer->lineHeight();
  uint lineMin    = startY() / lineHeight;
  uint lineMax    = (startY() + height()) / lineHeight;
  uint lineStart  = mHexBuffer->calculateLine( start );
  uint lineStop   = mHexBuffer->calculateLine( stop );

  lineMin = lineStart > lineMin ? lineStart : lineMin;
  lineMax = lineStop  < lineMax ? lineStop  : lineMax; 

  int y = lineMin * lineHeight - startY();
  for( ; lineMin <= lineMax; lineMin ++, y += lineHeight )
  {
    mHexBuffer->drawText( paint, lineMin, startX(), 0, width() );
    bitBlt( this, 0, y, &mTextBuffer, 0, 0, width(), lineHeight );
  }
  paint.end();
}


void CHexViewWidget::paintText( const QRect &r )
{
  QPainter paint;
  paint.begin( &mTextBuffer );
  paint.setFont( mHexBuffer->font() );

  int lineHeight = mHexBuffer->lineHeight();
  int docLine    = (startY() + r.y()) / lineHeight;
  int y          = docLine * lineHeight - startY();
  int yMax       = r.y() + r.height();
  int xMax       = r.x() + r.width();

  for( ; y < yMax; docLine ++, y += lineHeight )
  {
    mHexBuffer->drawText( paint, docLine, startX(), r.x(), xMax );
    bitBlt( this, r.x(), y, &mTextBuffer, r.x(), 0, r.width(), lineHeight );
  }
  paint.end();
}



void CHexViewWidget::paintTextOnCursor( bool finishWindow )
{
  QPainter paint;
  paint.begin( &mTextBuffer );
  paint.setFont( mHexBuffer->font() );

  int lineHeight = mHexBuffer->lineHeight();
  int docLine    = mHexBuffer->cursorLine();// - 1;
  int y          = docLine * lineHeight - startY();
  int yMax       = finishWindow == true ? height() : y+1;

  for( ; y < yMax; docLine ++, y += lineHeight )
  {
    mHexBuffer->drawText( paint, docLine, startX(), 0, width() );
    bitBlt( this, 0, y, &mTextBuffer, 0, 0, width(), lineHeight );
  }
  paint.end();
}





void CHexViewWidget::paintTextLine( uint docLine, int numLine )
{
  QPainter paint;
  paint.begin( &mTextBuffer );
  paint.setFont( mHexBuffer->font() );

  int lineHeight = mHexBuffer->lineHeight();
  int y          = docLine * lineHeight - startY();
  int yMax       = mEditMode == CHexBuffer::edit_insert || numLine <= 0 ? 
    height() : y + numLine * lineHeight;

  for( ; y < yMax; docLine ++, y += lineHeight )
  {
    mHexBuffer->drawText( paint, docLine, startX(), 0, width() );
    bitBlt( this, 0, y, &mTextBuffer, 0, 0, width(), lineHeight );
  }
  paint.end();
}




void CHexViewWidget::paintCursor( int cursorMode )
{
  QPainter paint;
  paint.begin( &mTextBuffer );
  paint.setFont( mHexBuffer->font() );

  int lineHeight = mHexBuffer->lineHeight();

  if( cursorMode == CHexBuffer::cursor_prev )
  {
    int line = mHexBuffer->prevCursorLine();
    int y    = line * lineHeight - startY();

    SCursorPosition p;
    mHexBuffer->prevCursor( CHexBuffer::edit_primary, p );
    mHexBuffer->drawText( paint, line, startX(), p.x, p.x + p.w );
    bitBlt( this, p.x, y, &mTextBuffer, p.x, 0, p.w, lineHeight );

    mHexBuffer->prevCursor( CHexBuffer::edit_secondary, p );
    mHexBuffer->drawText( paint, line, startX(), p.x, p.x + p.w );
    bitBlt( this, p.x, y, &mTextBuffer, p.x, 0, p.w, lineHeight );
  }
  else
  {
    int line = mHexBuffer->cursorLine();
    int y    = line * lineHeight - startY();

    SCursorPosition p;
    mHexBuffer->currCursor( CHexBuffer::edit_primary, p );
    mHexBuffer->drawText( paint, line, startX(), p.x, p.x + p.w );
    bitBlt( this, p.x, y, &mTextBuffer, p.x, 0, p.w, lineHeight );

    mHexBuffer->currCursor( CHexBuffer::edit_secondary, p );
    mHexBuffer->drawText( paint, line, startX(), p.x, p.x + p.w );
    bitBlt( this, p.x, y, &mTextBuffer, p.x, 0, p.w, lineHeight );
  }

  paint.end();
}



void CHexViewWidget::updateCursor( SCursorConfig &cc, bool always, 
				   bool touchSelection )
{
  if( mHexBuffer->cursorChanged() == false && always == false )
  {
    return;
  }
  
  //
  // Make blinking (and perhaps invisible) cursor visible
  //
  enableCursorTimer( true );

  //
  // Clear cursor at old location
  //
  paintCursor( CHexBuffer::cursor_prev );

  //
  // Compute the new position of the vertical scroll bar.
  //
  int position;
  if( cc.controlButton() == true )
  {
    //
    // The cursor should stay fixed (if possible) in the window while 
    // the text is scrolled (e.g., PageUp/Down behavior). The position 
    // of the vertical scrollbar must change just as much as the cursor 
    // has changed in the vertical direction.
    //
    position = mHexBuffer->cursorFixedPosition( startY(), height() );
    changeYPos( position );
  }
  else
  { 
    position = mHexBuffer->cursorChangePosition( startY(), height() );
    changeYPos( position );
  }

  //
  // Paint cursor at new location and update the vertical scroll bar.
  //
  paintCursor( CHexBuffer::cursor_curr );
  mVertScroll->blockSignals( true );
  mVertScroll->setValue( position );
  mVertScroll->blockSignals( false );

  if( touchSelection == true )
  {
    setSelection( mHexBuffer->cursorOffset(), cc.removeSelection() );
  }
  emit cursorChanged( mHexBuffer->cursorState() );
}



void CHexViewWidget::stepEditArea( void )
{
  bool success = mHexBuffer->stepActiveEditArea();
  if( success == false )
  {
    return;
  }

  SCursorConfig cc;
  updateCursor( cc, true );
  paintTextOnCursor( false ); // This fix a redrawing shortcoming
}


void CHexViewWidget::cursorStep( SCursorConfig &cc, uint stepSize )
{
  mHexBuffer->cursorStep( stepSize, cc.altButton() ? false : true, true );
  cc.emulateControlButton( false );
  updateCursor( cc );
}



void CHexViewWidget::cursorLeft( SCursorConfig &cc )
{
  bool cellLevel = cc.controlButton();
  cc.emulateControlButton( false );
  mHexBuffer->cursorLeft( cellLevel );
  updateCursor( cc, cellLevel );
}


void CHexViewWidget::cursorRight( SCursorConfig &cc )
{
  bool cellLevel = cc.controlButton();
  cc.emulateControlButton( false );
  mHexBuffer->cursorRight( cellLevel );
  updateCursor( cc, cellLevel );
}



void CHexViewWidget::cursorUp( SCursorConfig &cc )
{
  mHexBuffer->cursorUp( 1 );
  updateCursor( cc );
}


void CHexViewWidget::cursorDown( SCursorConfig &cc )
{
  mHexBuffer->cursorDown( 1 );
  updateCursor( cc );
}


void CHexViewWidget::cursorHome( SCursorConfig &cc )
{
  mHexBuffer->cursorHome( cc.controlButton() );
  updateCursor( cc );
}


void CHexViewWidget::cursorEnd( SCursorConfig &cc )
{
  mHexBuffer->cursorEnd( cc.controlButton() );
  updateCursor( cc );
}


void CHexViewWidget::cursorPageDown( SCursorConfig &cc )
{
  mHexBuffer->cursorDown( height() / mHexBuffer->lineHeight() );
  cc.emulateControlButton( true );
  updateCursor( cc );
}


void CHexViewWidget::cursorPageUp( SCursorConfig &cc )
{
  mHexBuffer->cursorUp( height() / mHexBuffer->lineHeight() );
  cc.emulateControlButton( true );
  updateCursor( cc );
}


void CHexViewWidget::cursorInsert( SCursorConfig &cc )
{
  setEditMode( mEditMode == CHexBuffer::edit_insert ? 
	       CHexBuffer::edit_overwrite : CHexBuffer::edit_insert );
}


void CHexViewWidget::cursorDelete( SCursorConfig &cc )
{
  int numLine = mHexBuffer->numLines();

  bool success = mHexBuffer->removeAtCursor( false );
  if( success == false )
  {
    return;
  }

  bool s = numLine == mHexBuffer->numLines() ? false : true;
  updateWindow( s, true );
  //updateWindow( false, true );
}


void CHexViewWidget::cursorBackspace( SCursorConfig &cc )
{
  int numLine = mHexBuffer->numLines();

  bool success = mHexBuffer->removeAtCursor( true );
  if( success == false )
  {
    return;
  }

  bool s = numLine == mHexBuffer->numLines() ? false : true;
  updateWindow( s, true );
  //updateWindow( false, true );
}


void CHexViewWidget::cursorInput( int key )
{
  if( key > 255 )
  {
    return;
  }

  uint cursorLine = mHexBuffer->cursorLine();
  
  bool success = mHexBuffer->inputAtCursor( key );
  if( success == false )
  {
    return;
  }

  updateWindow( cursorLine );
}


void CHexViewWidget::setEditMode( CHexBuffer::EEditMode mode )
{
  mEditMode = mode;
  mHexBuffer->setEditMode( mEditMode );
  enableCursorTimer( true );

  //
  // This will redraw the current line (which contains the cursor)
  //
  paintTextOnCursor( false );

  emit editMode( mEditMode );
}


